"""
    classes for Computer, devices, programs, files and some
    is* predicates operating on them
"""
from qt import *
import os, pickle, mimetypes
import config, persistence
from util import *

class Computer:
    def __init__(self, sortingCriterion, usageCount,
                 lastUsageTime, devicesLocation, localRootDirectory):
        self.icon = QPixmap(config.imgDir + "computer.png")
        self.sortingCriterion = sortingCriterion
        self.usageCount = usageCount
        self.lastUsageTime = lastUsageTime
        self.path = "all/computer"
        self.children = [devicesLocation, localRootDirectory]
    def toBriefString(self):
        return "computer"
    def toLongString(self):
        return self.path
    def persistentID(self):
        return ("Computer", self.path)
    def stringForPopupMenu(self):
        return self.toLongString()
    def containedObjects(self):
        return self.children
    def getLastUsageTime(self, hash):
        return self.lastUsageTime
    def setLastUsageTime(self, t, hash):
        self.lastUsageTime = t
    def increaseUsageCount(self, hash):
        self.usageCount += 1
    def getSortingCriterion(self, hash):
        return self.sortingCriterion
    def setSortingCriterion(self, crit, hash):
        self.sortingCriterion = crit
        return hash
    def getUsageCount(self, hash):
        return self.usageCount

class device:
    def __init__(self, name, icon, lastUsageTime, usageCount, path = 'all/computer/devices'):
        self.name = name
        self.lastUsageTime  = lastUsageTime
        self.usageCount = usageCount
        self.icon = icon
        self.isMonitor  = False
        self.canBurnCDs = False
        self.canBurnDVDs = False
        self.canReadCDs = False
        self.canReadDVDs = False
        self.isPrinter = False
        self.isScanner = False
        self.containsAFileSystem = False
        self.canBeEjected = False
        self.canBeMounted = False
        self.isSoundCard = False
        self.isMouse = False
        self.isKeyboard = False
        self.isDigitalCamera = False
        self.isHardDisk = False
        self.path = path
    def getLastUsageTime(self, hash):
        return self.lastUsageTime
    def setLastUsageTime(self, t, hash):
        self.lastUsageTime = t
    def toBriefString(self):
        return self.name
    def toLongString(self):
        return self.path +'/' + self.name
    def stringForPopupMenu(self):
        return self.toLongString()
    def persistentID(self):
        return ("device", self.path, self.name)
    def increaseUsageCount(self, hash):
        self.usageCount += 1
    def getUsageCount(self, hash):
        return self.usageCount

class program:
    def __init__(self, name, shellName, icon):
        self.name = name
        self.shellName = shellName
        self.icon = QPixmap(config.imgDir + icon)
        (uc , l) = persistence.loadVariableFromFile(config.dataDir + self.name, default = (0, None))
        self.usageCount = uc
        self.lastUsageTime = l
    def toBriefString(self):
        return self.name
    def toLongString(self):
        return self.name
    def persistentID(self):
        return ("program", self.name)
    def stringForPopupMenu(self):
        return self.name
    def getLastUsageTime(self, hash):
        return self.lastUsageTime
    def setLastUsageTime(self, t, hash):
        self.lastUsageTime = t
    def increaseUsageCount(self, hash):
        self.usageCount += 1
    def getUsageCount(self, hash):
        return self.usageCount
    def saveState(self): 
        f = file(config.dataDir + self.name, 'w')
        p = (self.usageCount, self.lastUsageTime)
        pickle.dump(p, f)
        f.close()

class LocalFile:
    def __init__(self, path):
        self.path = path
        ex = ext(self)
        if ex.endswith('~'):
            ex2 = ex[1:-1]
        else:
            ex2 = ex[1:]
        ic =  config.imgDir + config.typeToIcon.get(ex2, "generic.png")
        self.icon = QPixmap(ic)
    def getLastUsageTime(self, hash):
        return hash.get(self.path, None)
    def setLastUsageTime(self, t, hash):
        hash[self.path] = t
    def toBriefString(self):
        return os.path.basename(self.path)
    def toLongString(self):
        return self.path
    def persistentID(self):
        return ("LocalFile", self.path)
    def stringForPopupMenu(self):
        return self.toLongString()
    def getUsageCount(self, hash):
        return hash.get(self.path, 0)
    def increaseUsageCount(self, hash):
        count = hash.get(self.path, None)
        if count == None:
            hash[self.path] = 1
        else:
            hash[self.path] = count +1

# Virtual Locations are 'all', 'all/programs', 'all/computer',
# 'all/computer/devices', and subdevices like 'hard disk 1'.
class VirtualLocation:
    def __init__(self, path, children, icon, usageCount, sortingCriterion, lastUsageTime):
        self.path = path
        self.children = children
        self.icon = icon
        self.usageCount = usageCount
        self.lastUsageTime = lastUsageTime
        self.sortingCriterion = sortingCriterion
    def getLastUsageTime(self, hash):
        return self.lastUsageTime
    def setLastUsageTime(self, t, hash):
        self.lastUsageTime = t
    def containedObjects(self):
        return self.children
    def toBriefString(self):
        if self.path == '/':
            return "files"
        else:
            return os.path.basename(self.path)
    def stringForPopupMenu(self):
        return self.toLongString()
    def toLongString(self):
        return self.path
    def persistentID(self):
        return ("VirtualLocation", self.path)
    def increaseUsageCount(self, hash):
        self.usageCount += 1
    def getSortingCriterion(self, hash):
        return self.sortingCriterion
    def setSortingCriterion(self, crit, hash):
        self.sortingCriterion = crit
        return hash
    def getUsageCount(self, hash):
        return self.usageCount

# local directories are the directories on the file system, starting
# from /
class LocalFileSystemDirectory:
    def __init__(self, path):
        self.path = path
        assert os.path.isdir(path)
        if path == '/':
            self.icon = QPixmap(config.imgDir + "file-manager.png")
        else:
            self.icon = QPixmap(config.imgDir + "folder.png")
    def getLastUsageTime(self, hash):
        return hash.get(self.path, None)
    def setLastUsageTime(self, t, hash):
        hash[self.path] = t
    def getUsageCount(self, hash):
        return hash.get(self.path, 0)
    def toBriefString(self):
        if self.path == '/':
            return "files"
        else:
            return os.path.basename(self.path)
    def getSortingCriterion(self, hash):
        return hash.get(self.path, "Recent usage") # don't do _("Recent usage"))!!
    def toLongString(self):
        return self.path
    def persistentID(self):
        return ("LocalFileSystemDirectory", self.path)
    def stringForPopupMenu(self):
        return self.toLongString()
    def increaseUsageCount(self, hash):
        count = hash.get(self.path, None)
        if count == None:
            hash[self.path] = 1
        else:
            hash[self.path] = count +1
    def containedObjects(self):
        objs = [addSlashIfNeeded(self.path) + x for x in os.listdir(self.path)]
        dirs = [ LocalFileSystemDirectory(o) for o in objs if os.path.isdir(o)]
        files = [ LocalFile(o) for o in objs if os.path.isfile(o)]
        return dirs + files
    def setSortingCriterion(self, crit, hash):
        hash[self.path] = crit
        return hash


class req:
    def __init__(self, pred, name):
        self.pred = pred
        self.name = name

class verb:
    def __init__(self, name, requirementsList, iconFileName, keepLocationsVisibleWhenSelected,
             panelsToBeOpenedAutomatically, isGeneric):
        self.name = name
        self.isGeneric = isGeneric
        self.icon = QPixmap(config.imgDir + iconFileName)
        self.requirementsList = requirementsList
        (self.usageCount, self.lastUsageTime) = persistence.loadVariableFromFile(config.dataDir +'verb-'+self.name, default = (0, None))
        self.keepLocationsVisibleWhenSelected = keepLocationsVisibleWhenSelected
        self.panelsToBeOpenedAutomatically = panelsToBeOpenedAutomatically

    def isVisible(self, filterQuickText, selectedVerbs, selectedObjects):
        def matchesFilter(filter, v):
            s = filter['verb'].text()
            s2 = str(s)
            return v.name.lower().find(s2.lower()) != -1
        
        if selectedVerbs == []:
            return (matchesFilter(filterQuickText, self) and 
                    forall(selectedObjects,
                           lambda o: exists(self.requirementsList,
                                            lambda p:p.pred(o))))
                    # means: forall o in selectedObjects exists p in
                    # self.requirementsList such that p.pred(o).
        else:
            return self in selectedVerbs
    def stringForPopupMenu(self):
        return self.name
    def saveState(self):
        f = file(config.dataDir + 'verb-'+self.name, 'w')
        pickle.dump((self.usageCount, self.lastUsageTime), f)
        f.close()

def isLocation(l):
    return (  isinstance(l, VirtualLocation) or
              isinstance(l, LocalFileSystemDirectory) or
              isinstance(l, Computer))

def isFileWithExtension(e, f):
    if isinstance(f, LocalFile):
        return e == ext(f)
    else:
        return False

def isProgramSupporting(x, verb):
    return isinstance(x, program) and  (verb, x.name) in config.supports

def isFile(f):
    return isinstance(f, LocalFile) or isinstance(f, LocalFileSystemDirectory)
def isAudioFile(f):
    return isinstance(f, LocalFile) and  ext(f) in ['.ogg', '.ogm', '.mp3', '.mpc', '.flac', '.ape', '.rm', '.wma', '.wmv', '.wav']
def isSubtitleFile(f):
    return isinstance(f, LocalFile) and  ext(f) in ['.sub', '.srt']
def isHTMLFile(f):
    return isinstance(f, LocalFile) and  ext(f) in ['.html', '.xml', '.htm']
def isCalendarFile(f):
    return isinstance(f, LocalFile) and  ext(f) in ['.ics']
def isPictureFile(f):
    if isinstance(f, LocalFile):
        (t, enc) = mimetypes.guess_type(f.path)
        return   t != None and t.startswith('image/')
    return False



def isVideoFile(f):
    if isinstance(f, LocalFile):
        if  ext(f) in ['.avi', '.mpeg', '.mpg', '.asf', '.ogg', '.ogm', '.rm', '.wmv']:
            return True
        else:
            (t, enc) = mimetypes.guess_type(f.path)
            return   t != None and t.startswith('video/')
    return False

def isCompressedFile(f):
    return isinstance(f, LocalFile) and ext(f) in ['.zip', '.rar', '.tgz', '.gz','.tar','.arj','.ace','.lha','.bz2']

def isAudioFile(f):
    if isinstance(f, LocalFile):
        if  ext(f) in  ['.ogg', '.ogm', '.mp3', '.mpc', '.flac',
                        '.ape', '.rm', '.wma', '.wmv', '.wav']:
            return True
        else:
            (t, enc) = mimetypes.guess_type(f.path)
            return   t != None and t.startswith('audio/')
    return False
            
           
def isPlainTextFile(f):
    if isinstance(f, LocalFile):
        if ext(f) in ['.txt', '.spec', '.tex', '.srt', '.sub','.pl',
                      '.pas', '.cc', '.cpp', '.h','.hpp', '.c',
                      '.tex', '.sub', '.htm','.php', '.htm', '.html']:
            return True
        else:
            (t, enc) = mimetypes.guess_type(f.path)
            return   t != None and t.startswith('text/') 
    else:
        return False
        

