#!/usr/bin/env python
# -*- coding: iso8859-1 -*-
"""
   Logical Desktop main module

   @license: GNU GPL, see COPYING for details
   @copyright: 2004, Maurizio Colucci
"""

import sys, os,  datetime, pickle, shutil

import config, version, i18n, config_programs
_ = i18n._

from classes import *
from util import *
from qt import QMainWindow, QMessageBox
from qt_widgets import *
#from  kdecore import *
#from  kdeui import *

# terminology: a "predicate" is a boolean function, i.e. a function
# that returns true or false. A "unary" predicate is a boolean
# function that takes exactly one argument.

# BEGIN declare the global lists and dictionaries. These structures
# can be seen as the database used by the program. The content of
# these structures changes over time.
# TODO do this without using lots of global variables (and also
# using function parameters (== local vars) of the same name)
recentFilters = {}
recentLocations = []
favouriteLocations = []

objectUsageCount = {}
timeTheObjectWasLastUsed = {}

visibleLocationInPanel = {}
sortingCriterionForLocation = {}

visibleVerbs = []
selectedVerbs = []
selectedObjects = []

devices = []
devicesExceptComputer = []
programs = []
verbs = []

selectedItemsPopupMenu = None
bookmarksButton = None

locationButton = {}

listBox = {}
sortBox = {}

filterBox = {}
filterButton = {}
filterMenu = {}
filterQuickText = {}
filterClearButton = {}

selectedLabel = None
deselectAllButton = deselectOneItemButton = None
recentLocationsButton = executeButton = None
closeButton = {}

hBox = {}
icon = {}

favouriteLocPopupMenu = recentLocPopupMenu = None
executeToolbar = locationToolbar = None

localRootDirectory = None

myComputer = myPrinter = myHardDisk1 = MyHardDisk2 = MyMonitor = MyMonitor2 = None
MyKeyboard = MySoundCard = MyMouse = dvdReader = cdBurner = None
devicesLocation = programsLocation = allLocation = None
# END

# Qt stuff
# app renamed to applic to be easier searchable (see "append", "apply", etc.)
applic = mainWindow = None

gDefaultMessageWhenNothingSelected = _("Please select a verb, a file, a program or a device.")

        
def point_to_string(p):
    if p is None:
        return None
    else:
        return p.x, p.y


class MyMainWindow(QMainWindow):
    # used on qt. 
    def closeEvent(self, ev):
        global recentLocations
        recentLocations = addCurrentlyVisibleFileLocationToRecentLocations(recentLocations, visibleLocationInPanel)
        QMainWindow.closeEvent(self, ev)
        #return True # accept the closing
    
    # used by kde, useless in the qt version
    #    def queryClose(self):
    #        global recentLocations
    #        recentLocations = addCurrentlyVisibleFileLocationToRecentLocations(recentLocations, visibleLocationInPanel)
    #        return True # accept the closing

def verifyThatNoVerbContainsInvalidCharacters():
    v = find(verbs, lambda v:v.name.find('/') != -1)
    if v != None:
        print 'Integrity error: the verb "%s" contains the character "/"' % v.name
        assert False

def clearFileFiltersAndUpdate():
    clearAux('file')
def clearVerbFiltersAndUpdate():
    clearAux('verb')
def clearProgramFiltersAndUpdate():
    clearAux('program')
def clearDeviceFiltersAndUpdate():
    clearAux('device')

def clearAux(k):
    global visibleVerbs
    filterQuickText[k].clear()
    visibleVerbs = redrawAllPanels(verbs, selectedVerbs, selectedObjects,
                       programsLocation, devicesLocation,
                       visibleLocationInPanel, timeTheObjectWasLastUsed, objectUsageCount,
                       selectedItemsPopupMenu,
                       centeringCommand = None)

def onFilterKeyTimer(widget):
    applyFilter()
    
def onEnterPressedOnFilter():
    applyFilter()
    
def applyFilter():
    global visibleVerbs
    visibleVerbs = redrawAllPanels(
        verbs, selectedVerbs, selectedObjects,
        programsLocation, devicesLocation,
        visibleLocationInPanel, timeTheObjectWasLastUsed,
        objectUsageCount,
        selectedItemsPopupMenu,
        centeringCommand = None)
    # BEGIN rebuild recentFilters[k] for all k
    for k in ['verb', 'file', 'program', 'device']:
        t = str(filterQuickText[k].text())
        if t:
            if t in recentFilters[k]:
                recentFilters[k].remove(t)
            recentFilters[k] = [t] + recentFilters[k]
        recentFilters[k] = (recentFilters[k])[:config.gMaxItemsInRecentLocMenu] # truncate
    # END
    rebuildFilterMenus()

def rebuildFilterMenus():
    for k in ['verb', 'file', 'program', 'device']:
        filterMenu[k].clear()
        l = recentFilters[k]
        l.reverse()
        for f in l:
            filterMenu[k].addItem(PopupMenuItem(f), QPixmap())
        l.reverse()

def onRecentFilterClicked(menu, filter):
    k = find(['verb', 'file', 'program', 'device'], lambda x: menu is filterMenu[x])
    filterQuickText[k].setText(filter.stringForPopupMenu())
    onEnterPressedOnFilter()

def equal(o, o2):
    return o.persistentID() == o2.persistentID()

         
def theObjectMatchesFilter(o):
    if isinstance(o, device) and o.path == 'all/computer/devices':
        k = 'device'
    elif isinstance(o, program):
        k = 'program'
    else:
        k = 'file'
    s = str(filterQuickText[k].text())
    return o.toBriefString().lower().find(s.lower()) != -1
        
# Intuition: an object o is visible if 1) o is a location, or 2) there
# is a visible verb that "requires" o, where "v requires o" means
# "exists p in v.requirementsList such that p(o)". Now let us put it
# all together:
def theObjectIsVisible(o, visibleVerbs):
    return (   theObjectMatchesFilter(o) and
               ( (   isLocation(o) and
                     exists(visibleVerbs,
                            lambda v:v.keepLocationsVisibleWhenSelected)) or
                 # exists v in visibleVerbs such that exists p in
                 # v.requirementsList such that p.pred(o) is true.
                 exists(visibleVerbs,
                        lambda v: exists(v.requirementsList, lambda p:p.pred(o)))))

def correctPanelForDisplayingTheContentOfTheLocation(l):
    if isinstance(l, VirtualLocation) and l.path == 'all/programs':
        return 'program'
    elif isinstance(l, VirtualLocation) and l.path == 'all/computer/devices':
        return 'device'
    else:
        return 'file'

        
def pressWrappingButtonIfTheObjectIsSelected(o, selectedObjects):
    k = correctPanelForDisplayingTheContentOfTheLocation(o)
    locationButton[k].setPressedState(pressed = exists(selectedObjects, lambda o2: equal(o,o2)))

def theLocationButtonIsClickable(l, visibleVerbs,  selectedObjects):
    return (   exists(selectedObjects,
                      lambda l2:equal(l, l2)) or
               # exists a visible verb v: exists p in
               # v.requirementsList: p.pred(l).
               exists(visibleVerbs,
                      lambda v:exists(v.requirementsList,
                                      lambda p:p.pred(l))))

        
def updateExecuteButton(visibleVerbs, selectedObjects, visibleObjects):
    if len(visibleVerbs) == 1:
        v = visibleVerbs[0]
        if  allTheRequiredObjectsAreSelectedOrObvious(v, selectedObjects, visibleObjects):
            executeButton.startFlashing(config.gFlashingTimeForButtons, resetOriginalColorsWhenFinished = True)
            executeToolbar.show()
            t = _('Execute "%s"') % v.name
            l = len([x for x in selectedObjects if isFile(x)])
            if l == 0:
                t2 = ""
            elif l == 1:
                t2 = _(" (with 1 file)")
            else:
                t2 = _(" (with %d files)") % l
            executeButton.setText(t + t2)
        else:
            executeToolbar.hide()
    else:
        executeToolbar.hide()

              
def allTheRequiredObjectsAreSelectedOrObvious(v, selectedObjects, visibleObjects):
    # allTheRequiredObjectsAreSelectedOrObvious(v) is true if 1)
    # forall p in v.requirementsList, exists o in selectedObjects such
    # that p.pred(o), or 2) if there is only one o in visibleObjects
    # such that p.pred(o).
    return forall(v.requirementsList,
                  lambda p: (   exists(selectedObjects, lambda o: p.pred(o)) or
                                len([o for o in visibleObjects if p.pred(o)]) == 1))

def isTheOnlyElementInListThatSatisfies(el, list, pred):
    return (   pred(el) and
               # forall o in list: o is equal to el OR pred(o) is
               # false.
               forall(list, lambda o: equal(o, el) or not pred(o)))


class action:
    def __init__(self, v, objs):
        self.verb = v
        self.objects = objs

def executeActionNow(a):
    if a.verb.name in [ 'Play audio',
                        'Edit plain text file',
                        'Edit SXW file',
                        'Edit DOC file',
                        'Preview images contained in location',
                        'View DVI file',
                        'Open terminal in directory',
                        'View PS file',
                        'Search directory',
                        'Open directory with Cervisia',
                        'View picture',
                        'Open with Konqueror',
                        'View contents of archive',
                        'View files contained in RPM',
                        'View PDF',
                        'Play movie', ]:
        simplyExecuteAllTheProgramsWithAllTheFiles(a.objects)

    elif a.verb.name == "Add to important locations":
        for o in a.objects:
            if not exists(favouriteLocations, lambda o2: equal(o,o2)):
                favouriteLocations.append(o)
        rebuildImportantLocationsMenu(favouriteLocations, allLocation)
        bookmarksButton.startFlashing(config.gFlashingTimeForButtons, resetOriginalColorsWhenFinished = True)

    elif a.verb.name in ["Enqueue audio for playing", "Enqueue video for playing"]:
        progs = [o for o in a.objects if isinstance(o, program)]
        files = [ o for o in a.objects if isFile(o)]
        files2 = reduce( lambda x, y: x+y,
                         map(lambda f: '"%s"' % f.path, files),
                         "")
        for p in progs:
            s = "%s --enqueue %s &" % (p.shellName, files2)
            print s
            os.system(s)

    elif a.verb.name in ["Play movie with subtitle file"]:
        progs = [o for o in a.objects if isinstance(o, program)]
        videofiles = [ o for o in a.objects if isVideoFile(o)]
        subfiles = [ o for o in a.objects if isSubtitleFile(o)]
        if len(subfiles) != 1:
            QMessageBox.critical(mainWindow, _("Error"), _("There must be exactly one subtitle file selected."))
        elif len(progs) != 1:
            QMessageBox.critical(mainWindow, _("Error"), _("There must be exactly one program selected."))
        elif len(videofiles) != 1:
            QMessageBox.critical(mainWindow, _("Error"), _("There must be exactly one video file selected."))
        else:
            cmdLIne= '%s "%s" -sub "%s"' % (progs[0].shellName, videofiles[0].path, subfiles[0].path)
            print cmdLIne
            os.system(cmdLIne)
            
    elif a.verb.name in ["Play movie with subtitle file and TrueType font"]:
        progs = [o for o in a.objects if isinstance(o, program)]
        fonts = [o for o in a.objects if  isinstance(o, LocalFile) and ext(o)=='.ttf']
        videofiles = [ o for o in a.objects if isVideoFile(o)]
        subfiles = [ o for o in a.objects if isSubtitleFile(o)]
        if len(subfiles) != 1:
            QMessageBox.critical(mainWindow, _("Error"), _("There must be exactly one subtitle file selected."))
        if len(fonts) != 1:
            QMessageBox.critical(mainWindow, _("Error"), _("There must be exactly one TTF file."))
        elif len(progs) != 1:
            QMessageBox.critical(mainWindow, _("Error"), _("There must be exactly one program selected."))
        elif len(videofiles) != 1:
            QMessageBox.critical(mainWindow, _("Error"), _("There must be exactly one video file selected."))
        else:
            cmdLIne= '%s "%s" -sub "%s" -font "%s"' % (progs[0].shellName, videofiles[0].path, subfiles[0].path, fonts[0].path)
            print cmdLIne
            os.system(cmdLIne)

    elif a.verb.name == "Copy files to directory":
        if len(a.objects) < 2:
            QMessageBox.critical(mainWindow, _("Error"), _("You need to specify at least 2 objects."))
        else:
            if len([d for d in a.objects if os.path.isdir(d.path)]) == 1:
                d = find(a.objects, lambda d:os.path.isdir(d.path))
                files = [ x for x in a.objects if os.path.isfile(x.path)]
                try:
                    for f in files:
                        shutil.copy2(f.path, d.path)
                except IOError, (err, strerr):
                    QMessageBox.critical(mainWindow, _("Error"), strerr, QMessageBox.Ok)
            else:
                QMessageBox.information(mainWindow, _("Ambiguous"),
                                        _("The command is ambiguous. In later versions, Logical Desktop will ask you which is the destination directory."),
                                        QMessageBox.Ok)
        
    elif a.verb.name == "Move files to directory":
        if len(a.objects) < 2:
            QMessageBox.critical(mainWindow, _("Error"), _("You need to specify at least 2 objects."))
        else:
            if len([d for d in a.objects if os.path.isdir(d.path)]) == 1:
                d = find(a.objects, lambda d:os.path.isdir(d.path))
                files = [ x for x in a.objects if os.path.isfile(x.path)]
                try:
                    for f in files:
                        shutil.move(f.path, d.path)
                except IOError, (err, strerr):
                    QMessageBox.critical(mainWindow, _("Error"), strerr, QMessageBox.Ok)
            else:
                QMessageBox.information(mainWindow, _("Ambiguous"),
                                        _("The command is ambiguous. In later versions, Logical Desktop will ask you which is the destination directory."),
                                        QMessageBox.Ok)
        
    elif a.verb.name == "Delete file":
        if QMessageBox.critical(mainWindow,
                                _("Confirm deletion"),
                                _("Are you sure you want to delete %d files?") % len(a.objects),
                                QMessageBox.Ok, QMessageBox.Cancel) == QMessageBox.Ok:
            for f in a.objects:
                if os.path.isfile(f.path):
                    try:
                        os.remove(f.path)
                    except OSError, (errno, strerror):
                        QMessageBox.critical(mainWindow, _("Error"), strerror, QMessageBox.Ok)
                        
                elif os.path.isdir(f.path):
                    shutil.rmtree(f.path)
                else:
                    raise "not handled"
            selectedObjects = []
        
    elif a.verb.name == "Remove from important locations":
        if allLocation in a.objects:
            QMessageBox.critical(None, _("Error"), _('The "all" location cannot be removed.'))
        else:
            for o in a.objects:
                l = find(favouriteLocations, lambda o2: equal(o,o2))
                if l != None:
                    favouriteLocations.remove(l)
            rebuildImportantLocationsMenu(favouriteLocations, allLocation)
            bookmarksButton.startFlashing(config.gFlashingTimeForButtons, resetOriginalColorsWhenFinished = True)
    else:
        QMessageBox.critical(None, _("Error"), _('Verb "%s" not yet implemented.\n Please consider helping the author. :-)') % a.verb.name)

def simplyExecuteAllTheProgramsWithAllTheFiles(objs):
    assert not exists(objs, lambda o: isinstance(o, device))
    progs = [o for o in objs if isinstance(o, program)]
    files = [o for o in objs if isFile(o)]
    files2 = reduce( lambda x, y: x+ " " +y,
                     ['"%s"' % f.path for f in  files],
                     "")
    for p in progs:
        cmd = '%s %s &' % (p.shellName, files2)
        print cmd
        os.system(cmd)
    
def onExecuteButtonPressed():
    global visibleVerbs
    assert len(visibleVerbs) == 1
    v = visibleVerbs[0]
    openPanels = [ k for k in ['file', 'program', 'device'] if visibleLocationInPanel.get(k, None) != None]
    visibleLocations = [ visibleLocationInPanel[p] for p in openPanels]
    visibleObjects = sum( [loc.containedObjects() for loc in visibleLocations], []) + visibleLocations
    visibleNotSelected = [o for o in visibleObjects if not exists(selectedObjects, lambda o2: equal(o, o2))]
    
    inferredObjects = [o for o in visibleNotSelected if exists(v.requirementsList,
                                                               lambda p: isTheOnlyElementInListThatSatisfies(o, visibleObjects, p.pred))]
    allUsedObjects = inferredObjects + selectedObjects
    for o in allUsedObjects:
        o.increaseUsageCount(objectUsageCount)
        o.setLastUsageTime(datetime.datetime.now(), timeTheObjectWasLastUsed)
    v.usageCount += 1
    v.lastUsageTime = datetime.datetime.now()
    a = action(v, allUsedObjects)
    executeActionNow(a)
    visibleVerbs = redrawAllPanels(verbs, selectedVerbs, selectedObjects,
                                   programsLocation, devicesLocation,
                                   visibleLocationInPanel, timeTheObjectWasLastUsed, objectUsageCount,
                                   selectedItemsPopupMenu,
                                   centeringCommand = None)

    
def onFileCloseButtonPressed():
    k = 'file'
    listBox[k].hide()
    sortBox[k].hide()
    filterBox[k].hide()
    hBox[k].hide()
    del visibleLocationInPanel[k]
    
def onProgramCloseButtonPressed():
    k = 'program'
    listBox[k].hide()
    sortBox[k].hide()
    filterBox[k].hide()
    hBox[k].hide()
    del visibleLocationInPanel[k]
    
def onDeviceCloseButtonPressed():
    k = 'device'
    listBox[k].hide()
    sortBox[k].hide()
    filterBox[k].hide()
    hBox[k].hide()
    del visibleLocationInPanel[k]
    
def onDeselectAllPressed():
    global selectedObjects, selectedVerbs, visibleVerbs
    selectedObjects = []
    selectedVerbs = []
    selectedLabel.setText(gDefaultMessageWhenNothingSelected)
    executeToolbar.hide()
    visibleVerbs = redrawAllPanels(verbs, selectedVerbs, selectedObjects,
                                   programsLocation, devicesLocation,
                                   visibleLocationInPanel, timeTheObjectWasLastUsed, objectUsageCount,
                                   selectedItemsPopupMenu,
                                   centeringCommand = None)
    
def onLocationButtonPressed(but):
    global visibleVerbs
    k  = find(['program', 'device', 'file'], lambda x:locationButton[x] is but)
    assert visibleLocationInPanel[k] is but.data
    selectedObjects.append(but.data)
    deselectOneItemButton.startFlashing(config.gFlashingTimeForButtons, resetOriginalColorsWhenFinished = True)
    visibleVerbs = redrawAllPanels(verbs, selectedVerbs, selectedObjects,
                                   programsLocation, devicesLocation,
                                   visibleLocationInPanel, timeTheObjectWasLastUsed, objectUsageCount,
                                   selectedItemsPopupMenu,
                                   centeringCommand = None)
    
def onLocationButtonReleased(but):
    global visibleVerbs
    k  = find(['program', 'device', 'file'], lambda x:locationButton[x] is but)
    assert visibleLocationInPanel[k] is but.data
    selectedObjects.remove(but.data)
    deselectOneItemButton.startFlashing(config.gFlashingTimeForButtons, resetOriginalColorsWhenFinished = True)
    visibleVerbs = redrawAllPanels(verbs, selectedVerbs, selectedObjects,
                        programsLocation, devicesLocation,
                        visibleLocationInPanel, timeTheObjectWasLastUsed, objectUsageCount,
                        selectedItemsPopupMenu,
                        centeringCommand = None)
    
    
# callback :-(
def onEnterLocationButtonClicked(listBox, item):
    global recentLocations, visibleVerbs
    recentLocations = addCurrentlyVisibleFileLocationToRecentLocations(recentLocations, visibleLocationInPanel)
    loc = item.data
    k = correctPanelForDisplayingTheContentOfTheLocation(loc)
    filterQuickText[k].clear()
    openLocationIfNecessary(loc, visibleLocationInPanel)
    visibleLocationInPanel[k] = loc
    loc.increaseUsageCount(objectUsageCount)
    loc.setLastUsageTime(datetime.datetime.now(), timeTheObjectWasLastUsed)
    recentLocationsButton.startFlashing(config.gFlashingTimeForButtons, resetOriginalColorsWhenFinished = True)
    cr = _(loc.getSortingCriterion(sortingCriterionForLocation)) # sic! getSortingCriterion is an english string
    sortBox[k].setCurrentText(cr)
    # I need to redraw all the panels, not just the current one,
    # because entering a directory changes the visible objects,
    # which can change the help message and hide/show the
    # executeButton.
    visibleVerbs = redrawAllPanels(verbs, selectedVerbs, selectedObjects,
                                   programsLocation, devicesLocation,
                                   visibleLocationInPanel, timeTheObjectWasLastUsed, objectUsageCount,
                                   selectedItemsPopupMenu,
                                   centeringCommand = None)
    
def redrawLocation(l, selectedObjects, visibleVerbs,
                   objectUsageCount, timeTheObjectWasLastUsed,
                   centeringCommand):
    def addObjectToListWithProperSettings(o, selectedObjects, k, visibleVerbs):
        listBox[k].items.append(TextItem(text = o.toBriefString(),
                                         icon = o.icon,
                                         haveButton = isLocation(o),
                                         isSelectable =  theObjectIsSelectable(o, visibleVerbs),
                                         width = listBox[k].listArea.width(),
                                         pressed = exists(selectedObjects, lambda x:equal(x, o)),
                                         data = o))

    def splitAndSortObjects(k, objs, timeTheObjectWasLastUsed, objectUsageCount):
        if k == 'program':
            objs2 = [ o for o in objs if o.shellName in config_programs.configured_and_availab_prog_names]
            not_installed = [ o for o in objs if not o.shellName in config_programs.configured_and_availab_prog_names ]
            not_installed.sort(lambda o1, o2: cmp(o1.toBriefString().lower(),
                                                  o2.toBriefString().lower()))
            t3 = _('Not installed')
        else:
            objs2 = objs
            not_installed = []
            t3 = ''

        tx = sortBox[k].currentText()
        if  tx == _("Name"):
            objs2.sort(lambda o1, o2: cmp(o1.toBriefString().lower(),
                                         o2.toBriefString().lower()))
            return ([], objs2, not_installed, '', _('Sorted by name'), t3)
        elif tx == _('Recent usage'):
            usedObjs = [ x for x in objs2 if x.getLastUsageTime(timeTheObjectWasLastUsed) != None]
            unusedObjs = [x for x in objs2 if x.getLastUsageTime(timeTheObjectWasLastUsed) == None]
            unusedObjs.sort(lambda o1, o2: cmp(o1.toBriefString().lower(),
                                               o2.toBriefString().lower()))
            usedObjs.sort(lambda o1, o2: -cmp(o1.getLastUsageTime(timeTheObjectWasLastUsed),
                                              o2.getLastUsageTime(timeTheObjectWasLastUsed)))
            return (usedObjs, unusedObjs, not_installed,
                    _('Sorted by recent usage'), _('Never used, sorted by name'), t3)
        elif tx == _("Frequent usage"):
            usedObjs = [x for x in objs2 if x.getUsageCount(objectUsageCount) != 0]
            unusedObjs = [x for x in objs2 if x.getUsageCount(objectUsageCount) == 0]
            unusedObjs.sort(lambda o1, o2: cmp(o1.toBriefString().lower(),
                                               o2.toBriefString().lower()))
            usedObjs.sort(lambda o1, o2: - cmp(o1.getUsageCount(objectUsageCount),
                                               o2.getUsageCount(objectUsageCount)))
            return (usedObjs, unusedObjs, not_installed,
                    _('Sorted by frequent usage'), _('Never used, sorted by name'), t3)
        elif tx  == _("Date"):
            objs2.sort(lambda o1, o2: -cmp(os.path.getmtime(o1.path), os.path.getmtime(o2.path)))
            return ([], objs2,not_installed,
                    '', 'Most recently modified first', t3)
        elif tx == _("Size"):
            withSize = [o for o in objs2 if os.path.isfile(o.path)]
            withoutSize = [ o for o in objs2 if not os.path.isfile(o.path)]
            withSize.sort(lambda o1, o2: -cmp(os.path.getsize(o1.path), os.path.getsize(o2.path)))
            withoutSize.sort(lambda o1, o2: cmp(o1.toBriefString().lower(),
                                                o2.toBriefString().lower()))
            return (withSize, withoutSize, not_installed,
                    _('Biggest first'), _('files with no size'), t3)
        else:
            raise "error: the current text is invalid: " + tx + " " + sortBox[k].currentText()

    k = correctPanelForDisplayingTheContentOfTheLocation(l)
    assert l != None
    locationButton[k].setText(l.toLongString())
    locationButton[k].data = l
    icon[k].setPixmap(l.icon)
    pressWrappingButtonIfTheObjectIsSelected(l, selectedObjects)
    locationButton[k].setEnabled(
        theLocationButtonIsClickable(l,
                                     visibleVerbs,
                                     selectedObjects))
    listBox[k].clear()
    objs2 = [x for x in l.containedObjects() if theObjectIsVisible(x, visibleVerbs)]
    (o2a, o2b, o3b, ta, tb, tc) = splitAndSortObjects(k, objs2,
                                                      objectUsageCount = objectUsageCount,
                                                      timeTheObjectWasLastUsed = timeTheObjectWasLastUsed)
    if not ta == '' and not o2a == []:
        listBox[k].items.append(TextSeparator(ta, listBox[k].listArea.width()))
    for o in o2a:
        addObjectToListWithProperSettings(o, selectedObjects, k, visibleVerbs)
    if not tb == '' and not o2b == []:
        listBox[k].items.append(TextSeparator(tb, listBox[k].listArea.width()))
    for o in o2b:
        addObjectToListWithProperSettings(o, selectedObjects, k, visibleVerbs)
    if not tc == '' and not o3b == []:
        listBox[k].items.append(TextSeparator(tc, listBox[k].listArea.width()))
    for o in o3b:
        addObjectToListWithProperSettings(o, selectedObjects, k, visibleVerbs)
    if centeringCommand != None and centeringCommand.box is listBox[k]:
        listBox[k].centerItem(centeringCommand.itemText, centeringCommand.yoffset)
    listBox[k].update()
    return objs2 + [l]
    

# not all visible objects are also selectable. For example, suppose
# the verb "delete file" is selected. Then, the "/all/computer"
# location is visible (because you must be able to enter it) but not
# selectable (because it is not a file).
def theObjectIsSelectable(o, visibleVerbs):
    return ( not isLocation(o) or
             exists(visibleVerbs,
                    lambda v: exists(v.requirementsList, lambda p:p.pred(o))))


def redrawVerbs(verbs, selectedVerbs, selectedObjects, centeringCommand):
    def addVerbToListBox(v, selectedVerbs):
        listBox['verb'].items.append(TextItem(
            text = v.name, icon = v.icon,
            haveButton = False, isSelectable = True,
            width = listBox['verb'].listArea.width(),
            pressed = v in selectedVerbs,
            data = v))
    def splitAndSortVerbs(verbs):
        t = sortBox['verb'].currentText()
        if t == _("Name"):
            verbs.sort(lambda v1, v2: cmp(v1.name, v2.name))
            return ([], verbs, '', _('Sorted by name'))
        elif t == _('Recent usage'):
            recentVerbs = [v for v in verbs if v.lastUsageTime != None]
            nonRecentVerbs = [v for v in verbs if v.lastUsageTime is None]
            nonRecentVerbs.sort(lambda v1, v2: cmp(v1.name, v2.name))
            recentVerbs.sort(lambda v1, v2: -cmp(v1.lastUsageTime,  v2.lastUsageTime))
            return (recentVerbs, nonRecentVerbs, _('Sorted by recent usage'), _('Never used, sorted by name'))
        elif t == _("Frequent usage"):
            usedVerbs = [v for v in verbs if v.usageCount != 0]
            unusedVerbs  = [v for v in verbs if v.usageCount == 0]
            unusedVerbs.sort(lambda v1, v2: cmp(v1.name, v2.name))
            usedVerbs.sort(lambda v1, v2: -cmp(v1.usageCount , v2.usageCount))
            return (usedVerbs, unusedVerbs, _('Sorted by frequent usage'), _('Never used, sorted by name'))
        else:
            raise "unmanaged case"



    listBox['verb'].clear()
    visibleVerbs = [v for v in verbs if  v.isVisible(filterQuickText, selectedVerbs, selectedObjects)]
    if (selectedObjects != [] and
        forall(selectedObjects,
               lambda o: not isinstance(o, program) and not isinstance(o, device))):
        typ = [v for v in visibleVerbs if not v.isGeneric]
        atyp = [v for v in visibleVerbs if v.isGeneric]
        (typ1, typ2, t1, t2) = splitAndSortVerbs(typ)
        (atyp1, atyp2, a1, a2) = splitAndSortVerbs(atyp)
        typAll = typ1 + typ2
        atypAll = atyp1 + atyp2
        if typAll != []:
            listBox['verb'].items.append(TextSeparator('Typical', listBox['verb'].listArea.width()))
        for v in typAll:
            addVerbToListBox(v, selectedVerbs)
        if atypAll != [] and typAll != []:
            listBox['verb'].items.append(TextSeparator('Generic', listBox['verb'].listArea.width()))
        if atypAll != [] and typAll == []:
            listBox['verb'].items.append(TextSeparator(a1, listBox['verb'].listArea.width()))
        for v in atypAll:
            addVerbToListBox(v, selectedVerbs)
    else:

        vis1, vis2, text1, text2 = splitAndSortVerbs(visibleVerbs)
        if not text1 == '' and not vis1 == []:
            listBox['verb'].items.append(TextSeparator(text1, listBox['verb'].listArea.width()))

        for v in vis1:
            addVerbToListBox(v, selectedVerbs)
        if not text2 == '' and not vis2 == []:
            listBox['verb'].items.append(TextSeparator(text2, listBox['verb'].listArea.width()))
        for v in vis2:
            addVerbToListBox(v, selectedVerbs)
            
    listBox['verb'].update()
    return visibleVerbs


def addCurrentlyVisibleFileLocationToRecentLocations(recentLocations, visibleLocationInPanel):
    l = visibleLocationInPanel.get('file', None)
    if l is None:
        return recentLocations[:config.gMaxItemsInRecentLocMenu]
    else:
        ll = find(recentLocations, lambda l2: equal(l2, l))
        if not ll is None:
            recentLocations.remove(ll)
        recentLocations2 = [l] + recentLocations
        rebuildRecentLocationsMenu(recentLocations2)
        return recentLocations2[:config.gMaxItemsInRecentLocMenu]

def openLocationIfNecessary(l, visibleLocationInPanel):
    k = correctPanelForDisplayingTheContentOfTheLocation(l)
    curLoc = visibleLocationInPanel.get(k, None)
    if not curLoc is l:
        visibleLocationInPanel[k] = l
        cr = _(l.getSortingCriterion(sortingCriterionForLocation)) # sic!
        sortBox[k].setCurrentText(cr)
        hBox[k].show()
        listBox[k].show()
        sortBox[k].show()
        filterBox[k].show()
    listBox[k].startFlashing(config.gFlashingTimeForList, resetOriginalColorsWhenFinished = True)

def openTheLocationsRequiredByTheVerbIfNeeded(visibleVerbs, programsLocation,
                          devicesLocation, visibleLocationInPanel,
                          localRootDirectory):
    if len(visibleVerbs) > 1:
        return
    else:
        v = visibleVerbs[0]
        for k in v.panelsToBeOpenedAutomatically:
            if  k == 'device' and visibleLocationInPanel.get(k,None) == None:
                hBox[k].show()
                listBox[k].show()
                sortBox[k].show()
                filterBox[k].show()
                listBox[k].startFlashing(config.gFlashingTimeForList, resetOriginalColorsWhenFinished = True)
                visibleLocationInPanel[k] = devicesLocation
                cr = _(devicesLocation.getSortingCriterion(sortingCriterionForLocation)) # sic!
                sortBox[k].setCurrentText(cr)

            if  k == 'program' and visibleLocationInPanel.get(k, None) == None:
                hBox[k].show()
                listBox[k].show()
                sortBox[k].show()
                filterBox[k].show()
                listBox[k].startFlashing(config.gFlashingTimeForList, resetOriginalColorsWhenFinished = True)
                visibleLocationInPanel[k] = programsLocation
                cr = _(programsLocation.getSortingCriterion(sortingCriterionForLocation)) # sic!
                sortBox[k].setCurrentText(cr)
                
            if  k == 'file' and visibleLocationInPanel.get(k, None) == None:
                hBox[k].show()
                listBox[k].show()
                sortBox[k].show()
                filterBox[k].show()
                listBox[k].startFlashing(config.gFlashingTimeForList, resetOriginalColorsWhenFinished = True)
                visibleLocationInPanel[k] = localRootDirectory
                cr = _(localRootDirectory.getSortingCriterion(sortingCriterionForLocation)) # sic!
                sortBox[k].setCurrentText(cr)

def somethingIsSelected():
    #return not(selectedVerbs == [] and selectedObjects == [])
    return (selectedVerbs or selectedObjects)
            
def updateVisibilityOfSelectedItemsBar():
    if somethingIsSelected():
        deselectAllButton.setEnabled(True)
        deselectOneItemButton.setEnabled(True)
    else:
        deselectAllButton.setEnabled(False)
        deselectOneItemButton.setEnabled(False)
        selectedLabel.setText(gDefaultMessageWhenNothingSelected)
        for k in ['file', 'device', 'program']:
            closeButton[k].setEnabled(True)
        

def redrawAllPanels(verbs, selectedVerbs, selectedObjects,
            programsLocation, devicesLocation,
            visibleLocationInPanel, timeTheObjectWasLastUsed, objectUsageCount,
            selectedItemsPopupMenu,
            centeringCommand):
    visibleVerbs = redrawVerbs(verbs, selectedVerbs,
                   selectedObjects, centeringCommand)
    openTheLocationsRequiredByTheVerbIfNeeded(
        visibleVerbs, programsLocation,
        devicesLocation, visibleLocationInPanel,
        localRootDirectory)
    updateVisibilityOfSelectedItemsBar()
    visibleObjs = []
    for k in visibleLocationInPanel:
        visibleObjs += redrawLocation(visibleLocationInPanel[k],
                                      selectedObjects, visibleVerbs,
                                      objectUsageCount = objectUsageCount,
                                      timeTheObjectWasLastUsed = timeTheObjectWasLastUsed,
                                      centeringCommand = centeringCommand)
    updateExecuteButton(visibleVerbs, selectedObjects, visibleObjects = visibleObjs)
    redrawSelectedItemsPanel(selectedItemsPopupMenu,
                             visibleVerbs, visibleObjects = visibleObjs)
    return visibleVerbs

def redrawSelectedItemsPanel(selectedItemsPopupMenu, visibleVerbs, visibleObjects):
    # BEGIN calculate the text for the label
    assert len(visibleVerbs) != 0
    if len(visibleVerbs) > 1:
        selectedThings = selectedObjects + selectedVerbs
        l = len(selectedThings)
        if l == 1:
            t1 = _("There is 1 selected item, ")
            t2 = t1 + _("but the action is not yet complete.")
        elif l == 0:
            t2 = gDefaultMessageWhenNothingSelected
        else:
            t1 = _("There are %d selected items, ") % l
            t2 = t1 + _("but the action is not yet complete.")
        selectedLabel.stopFlashing(resetColor = True)
    else:
        v = visibleVerbs[0]
        if allTheRequiredObjectsAreSelectedOrObvious(v, selectedObjects, visibleObjects):
            t2 = _("You can now execute the action or select additional items.")
            selectedLabel.stopFlashing(resetColor = True)
        else:
            t2 = _("You still need to select")
            unsatisfiedPreds = [p for p in v.requirementsList if (not exists(selectedObjects, lambda o:p.pred(o)) and
                                                                  len([o for o in visibleObjects if p.pred(o)]) != 1)]
            if len(unsatisfiedPreds) == 1:
                t2 += " "
            else:
                t2 += ": "
            # XXX use ", ".join(list)+"."
            t2 += unsatisfiedPreds[0].name
            for p in unsatisfiedPreds[1:]:
                t2 += ", " + p.name
            t2+= '.'
            selectedLabel.startFlashing(config.gFlashingTimeForText, resetOriginalColorsWhenFinished = True)
    # END
        
    selectedLabel.setText(t2)
    selectedItemsPopupMenu.clear()
    if selectedVerbs != []:
        selectedItemsPopupMenu.addItem(selectedVerbs[0], selectedVerbs[0].icon)
    for o in selectedObjects:
        selectedItemsPopupMenu.addItem(o, o.icon)
        
def onSortBoxItemActivated(box, item):
    assert item in ["Recent usage", "Frequent usage", "Date", "Name", "Size"] # sic! don't change. Don't add _().
    global sortingCriterionForLocation
    if box is sortBox['verb']:
        redrawVerbs(verbs, selectedVerbs, selectedObjects, centeringCommand = None)
    else:
        k = find(['device', 'file', 'program'], lambda x: box is sortBox[x])
        loc = visibleLocationInPanel[k]
        # set an UNTRANSLATED (non-i18n) sorting mode to
        # sortingCriterionForLocation:
        sortingCriterionForLocation = loc.setSortingCriterion(item, sortingCriterionForLocation)
        redrawLocation(
            loc, selectedObjects, visibleVerbs, objectUsageCount = objectUsageCount,
            timeTheObjectWasLastUsed = timeTheObjectWasLastUsed,
            centeringCommand = None)

def onVerbClicked(box, item):
    global visibleVerbs
    v = item.data
    if v in selectedVerbs:
        selectedVerbs.remove(v)
    else:
        selectedVerbs.append(v)
    deselectOneItemButton.startFlashing(config.gFlashingTimeForButtons, resetOriginalColorsWhenFinished = True)
    visibleVerbs  = redrawAllPanels(verbs, selectedVerbs, selectedObjects,  programsLocation,
                    devicesLocation, visibleLocationInPanel, timeTheObjectWasLastUsed,
                    objectUsageCount, selectedItemsPopupMenu,
                    centeringCommand = None)

def onObjectItemClicked(box, item):
    global visibleVerbs
    x = find(selectedObjects, lambda o: equal(o, item.data))
    if not x is None:
        selectedObjects.remove(x)
    else:
        selectedObjects.append(item.data)
    centeringCommand = CenteringCommand(box, item.text, box.yOffsetOfItem(item))
    deselectOneItemButton.startFlashing(config.gFlashingTimeForButtons, resetOriginalColorsWhenFinished = True)
    visibleVerbs = redrawAllPanels(verbs, selectedVerbs, selectedObjects,
                        programsLocation, devicesLocation,
                       visibleLocationInPanel, timeTheObjectWasLastUsed, objectUsageCount,
                        selectedItemsPopupMenu,
                       centeringCommand)
    openTheLocationsRequiredByTheVerbIfNeeded(visibleVerbs, programsLocation, devicesLocation,
                          visibleLocationInPanel, localRootDirectory)
            
def onLocationSelectedFromPopupMenu(menu, l):
    # this is a callback, so it must have globals :-( 
    global recentLocations, visibleVerbs
    recentLocations = addCurrentlyVisibleFileLocationToRecentLocations(recentLocations, visibleLocationInPanel)
    l.increaseUsageCount(objectUsageCount)
    rebuildImportantLocationsMenu(favouriteLocations, allLocation)
    rebuildRecentLocationsMenu(recentLocations)
    openLocationIfNecessary(l, visibleLocationInPanel)
    k = correctPanelForDisplayingTheContentOfTheLocation(l)
    filterQuickText[k].clear()
    cr = _(l.getSortingCriterion(sortingCriterionForLocation)) # sic!
    sortBox[k].setCurrentText(cr)
    visibleVerbs = redrawAllPanels(verbs, selectedVerbs, selectedObjects,
                       programsLocation,
                       devicesLocation,
                       visibleLocationInPanel, timeTheObjectWasLastUsed, objectUsageCount,
                       selectedItemsPopupMenu,
                       centeringCommand = None)
    
def rebuildRecentLocationsMenu(recentLocations):
    recentLocPopupMenu.clear()
    for l in recentLocations:
        recentLocPopupMenu.addItem(l, l.icon)

def rebuildImportantLocationsMenu(favouriteLocations, allLocation):
    favouriteLocPopupMenu.clear()
    favouriteLocPopupMenu.addItem(allLocation, allLocation.icon)
    locs = [x for x in favouriteLocations if not (x is allLocation)]
    for l in locs:
        favouriteLocPopupMenu.addItem(l, l.icon)
    
def fromFileRepresentationToLocation(r):
    (ftype, fpath) = r
    if ftype == "VirtualLocation":
        if fpath == "all":
            return allLocation
        elif fpath == "all/programs":
            return programsLocation
        elif fpath == "all/computer/devices":
            return devicesLocation
        else:
            raise "unmanaged"
    elif ftype == "Computer":
        return myComputer
    else:
        assert ftype == "LocalFileSystemDirectory"
        if os.path.isdir(fpath):
            return LocalFileSystemDirectory(fpath)
        else:
            return None

# XXX is the stuff below still useful?

# verbRequiresAtLeastOneObjectSuchThat( 'Show/hide quick-search boxes',  isLogicalDesktop).
# verbRequiresAtLeastOneObjectSuchThat( 'Toggle scrolling buttons always visible',    isLogicalDesktop).
# verbRequiresAtLeastOneObjectSuchThat( 'Show/hide file sizes',  isLogicalDesktop).
# verbRequiresAtLeastOneObjectSuchThat( 'Show/hide file dates',  isLogicalDesktop).
# verbRequiresAtLeastOneObjectSuchThat( 'Quit',   isLogicalDesktop).
# verbRequiresAtLeastOneObjectSuchThat( 'Show/hide tooltips',  isLogicalDesktop).
# verbRequiresAtLeastOneObjectSuchThat( 'Show/hide log',   isLogicalDesktop).

##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Decompress APE to WAV',    ext(ape)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Decompress MP3 to WAV',    ext(mp3)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Decompress OGG to WAV',    ext(ogg)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Decompress OGM to WAV',    ext(ogm)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Compress WAV to MP3',    ext(wav)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Compress WAV to OGM',    ext(wav)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Compress WAV to FLAC',   ext(wav)).
##~ 
##~ 
##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Format partition',  diskPartition).
##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Share and download files over the internet',  supports('Share and download files over the internet')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Rip CD',  supports('Rip CD')).
##~ verbRequiresAtLeastOneObjectSuchThat('Rip CD',  cdDevice).
##~ 
##~ 
##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('View files contained in DEB',   ext(deb)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('View the dependencies of this RPM',   ext(rpm)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('View the dependencies of this DEB',   ext(deb)).
##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Mail files as attachments',   isFile).
##~ verbRequiresAtLeastOneObjectSuchThat('Mail files as attachments',  supports('Mail files as attachments')).
##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Extract audio track from OGM movie',   ogmFile).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Extract video track from OGM movie',   ogmFile).
##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Merge audio and video tracks',   videoFile).
##~ verbRequiresAtLeastOneObjectSuchThat('Merge audio and video tracks',   audioFile).
##~ verbRequiresAtLeastOneObjectSuchThat('Merge audio and video tracks',  supports('Merge audio and video tracks')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Install a new program',  supports('Install a new program')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Transfer files to and from an FTP site',  supports('Transfer files to and from an FTP site')).
##~ 
##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Open terminal',  supports('Open terminal')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Open terminal in directory',  supports('Open terminal in directory')).
##~ verbRequiresAtLeastOneObjectSuchThat('Open terminal in directory',   isADirectory).
##~ 

##~ verbRequiresAtLeastOneObjectSuchThat('Partition disk',  writableDisk).
##~ verbRequiresAtLeastOneObjectSuchThat('Partition disk',  supports('Partition disk')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Create thumbnails',  supports('Create thumbnails')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Set mouse speed',  mouseDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Set mouse speed',  supports('Set mouse speed')).
##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Set mouse cursor theme',  mouseDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Set mouse cursor theme',  supports('Set mouse cursor theme')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Read newsgroups',   supports('Read newsgroups')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Download a file from the internet with resume',  supports('Download a file from the internet with resume')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Take a snapshot of your screen',  supports('Take a snapshot of your screen')).  
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Setup K3B',   supports('Setup K3B')).
##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Rename multiple files at once',    isFile).
##~ verbRequiresAtLeastOneObjectSuchThat('Rename multiple files at once',   supports('Rename multiple files at once')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Install RPM',    ext(rpm)). 
##~ verbRequiresAtLeastOneObjectSuchThat('Install DEB',    ext(deb)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('View calendar and schedule',  supports('View calendar and schedule')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Eject disc',   ejectableDevice).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Create new text file',  supports('Create new text file')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Create new PDF file',  supports('Create new PDF file')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('View or terminate active processes',  supports('View or terminate active processes')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('View CPU usage',  supports('View CPU usage')).
##~ verbRequiresAtLeastOneObjectSuchThat('View memory usage',  supports('View memory usage')).
##~ 

##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Format disk',  writableDisk).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Play movie',    videoFile).
##~ verbRequiresAtLeastOneObjectSuchThat('Play movie',  supports('Play movie')).
##~ 
##~ ##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Erase CD-RW',  cdBurningDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Erase CD-RW',  supports('Erase CD-RW')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('View PDF',    ext(pdf)).
##~ verbRequiresAtLeastOneObjectSuchThat('View PDF',  supports('View PDF')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Burn from ISO',  supports('Burn from ISO')).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn from ISO',  cdBurningDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn from ISO',    ext(iso)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Burn audio CD',    audioFile).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn audio CD',  supports('Burn audio CD')).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn audio CD',  cdBurningDevice).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Burn video CD',    videoFile).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn video CD',  supports('Burn video CD')).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn video CD',  cdBurningDevice).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Print text file',    plainTextFile).
##~ verbRequiresAtLeastOneObjectSuchThat('Print text file',  supports('Print text file')).
##~ verbRequiresAtLeastOneObjectSuchThat('Print text file',  printerDevice).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Print PDF',    ext(pdf)).
##~ verbRequiresAtLeastOneObjectSuchThat('Print PDF',  supports('Print PDF')).
##~ verbRequiresAtLeastOneObjectSuchThat('Print PDF',  printerDevice).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Burn CD from BIN and CUE files',    ext(cue)).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn CD from BIN and CUE files',    ext(bin)).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn CD from BIN and CUE files',  cdBurningDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn CD from BIN and CUE files',  supports('Burn CD from BIN and CUE files')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Burn CD from audio and CUE files',    ext(cue)).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn CD from audio and CUE files',    audioFile).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn CD from audio and CUE files',  cdBurningDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Burn CD from audio and CUE files',  supports('Burn CD from audio and CUE files')).
##~ ##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Configure printer',  printerDevice).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Change keyboard layout',  keyboardDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Change keyboard layout',  supports('Change keyboard layout')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Change keyboard repeat speed',  keyboardDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Change keyboard repeat speed',  supports('Change keyboard repeat speed')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Set sound volume',  audioCardDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Set sound volume',  supports('Set sound volume')).
##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Shutdown',  computerDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Login as different user',  computerDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('Hybernate',  computerDevice).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Edit DOC file',    ext(doc)).
##~ verbRequiresAtLeastOneObjectSuchThat('Edit DOC file',  supports('Edit DOC file')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Edit SXW file',    ext(sxw)).
##~ verbRequiresAtLeastOneObjectSuchThat('Edit SXW file',  supports('Edit SXW file')).
##~ ##~ 
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Create new SXW file',  supports('Create new SXW file')).
##~ 
##~ ##~ verbRequiresAtLeastOneObjectSuchThat('Edit picture',  supports('Edit picture')).
##~ verbRequiresAtLeastOneObjectSuchThat('Edit picture',    pictureFile).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Print picture',  supports('Print picture')).
##~ verbRequiresAtLeastOneObjectSuchThat('Print picture',    pictureFile).
##~ verbRequiresAtLeastOneObjectSuchThat('Print picture',  printerDevice).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('View picture',  supports('View picture')).
##~ verbRequiresAtLeastOneObjectSuchThat('View picture',    pictureFile).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Change resolution',  monitorDevice).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Calculate size',    isFile).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('View print queue',  printerDevice).
##~ verbRequiresAtLeastOneObjectSuchThat('View print queue',  supports('View print queue')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Restore file from trashcan',  supports('Restore file from trashcan')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Empty trashcan',  supports('Empty trashcan')).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Move to trashcan',    isFile).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Edit TM file',  supports('Edit TM file')).
##~ verbRequiresAtLeastOneObjectSuchThat('Edit TM file',    ext(tm)).
##~ 
##~ verbRequiresAtLeastOneObjectSuchThat('Edit 3D picture',  supports('Edit 3D picture')).
##~ verbRequiresAtLeastOneObjectSuchThat('Edit 3D picture',   ext(blend)).

def makeprograms():
    for p in config.programs:
        programs.append(program(*p))

def makeverbs():
    for d in config.verbs:
        verbs.append(verb(**d))

def makedevices():
    global localRootDirectory
    global myComputer, myPrinter, myHardDisk1, MyHardDisk2, MyMonitor, MyMonitor2
    global MyKeyboard, MySoundCard, MyMouse, dvdReader, cdBurner
    global devices, devicesLocation, programsLocation, allLocation
    localRootDirectory = LocalFileSystemDirectory('/')

    # BEGIN create some fake devices for now. Later we will add device detection.
    myPrinter = device("HP printer", QPixmap(config.imgDir + "printer.png"),
                       lastUsageTime = None, usageCount = 0)
    myPrinter.isPrinter = True
    devicesExceptComputer.append(myPrinter)

    myHardDisk1 = device("Primary hard disk", QPixmap(config.imgDir + "devices.png"),
                         lastUsageTime = None, usageCount = 0)
    myHardDisk1.isHardDisk  = True
    myHardDisk1.containsAFileSystem = True
    devicesExceptComputer.append(myHardDisk1)

    myHardDisk2 = device("Secondary hard disk", QPixmap(config.imgDir + "devices.png"),
                         lastUsageTime = None, usageCount = 0)
    myHardDisk2.isHardDisk  = True
    myHardDisk2.containsAFileSystem = True
    myHardDisk2.canBeMounted  = True
    devicesExceptComputer.append(myHardDisk2)

    myMonitor = device("Sony Monitor", QPixmap(config.imgDir + "monitor.png"),
                       lastUsageTime = None, usageCount = 0)
    myMonitor.isMonitor = True
    devicesExceptComputer.append(myMonitor)

    myMonitor2 = device("LG Monitor", QPixmap(config.imgDir + "monitor.png"),
                        lastUsageTime = None, usageCount = 0)
    myMonitor2.isMonitor = True
    devicesExceptComputer.append(myMonitor2)

    myKeyboard = device("Logitech keyboard", QPixmap(config.imgDir + "keyboard.png"),
                        lastUsageTime = None, usageCount  = 0)
    myKeyboard.isKeyboard = True
    devicesExceptComputer.append(myKeyboard)

    mySoundCard = device("SoundBlaster 16", QPixmap(config.imgDir + "audio.png"),
                         lastUsageTime = None, usageCount = 0)
    mySoundCard.isSoundCard = True
    devicesExceptComputer.append(mySoundCard)

    myMouse = device("Logitech mouse", QPixmap(config.imgDir + "mouse.png"),
                     lastUsageTime = None, usageCount  = 0)
    myMouse.isMouse = True
    devicesExceptComputer.append(myMouse)

    dvdReader = device("Plextor DVD reader", QPixmap(config.imgDir + "cdrom.png"),
                       lastUsageTime = None, usageCount = 0)
    dvdReader.canReadDVDs = True
    dvdReader.canBeEjected = True
    dvdReader.canReadCDs = True
    dvdReader.containsAFileSystem = True
    devicesExceptComputer.append(dvdReader)

    cdBurner = device("LG CD burner", QPixmap(config.imgDir + "cdrom.png"),
                      lastUsageTime = None, usageCount = 0)
    cdBurner.canBurnCDs = True
    cdBurner.canBeEjected = True
    cdBurner.canReadCDs = True
    cdBurner.containsAFileSystem = True
    devicesExceptComputer.append(cdBurner)
    # END

    devicesLocation = VirtualLocation('all/computer/devices', devicesExceptComputer,
                                      QPixmap(config.imgDir + "devices.png"), usageCount = 0,
                                      sortingCriterion = _('Recent usage'), lastUsageTime = None)

    myComputer = Computer(sortingCriterion = _('Recent usage'),
                          usageCount  = 0,
                          lastUsageTime= None, devicesLocation = devicesLocation,
                          localRootDirectory = localRootDirectory)

    devices = devicesExceptComputer + [myComputer]

    programsLocation = VirtualLocation('all/programs', programs, QPixmap(config.imgDir + "bulb.png"),
                                       usageCount = 0,
                                       sortingCriterion = _('Recent usage'),
                                       lastUsageTime = None)

    allLocation = VirtualLocation('all', [myComputer, programsLocation],
                                  QPixmap(config.imgDir + "internet.png"), usageCount = 0,
                                  sortingCriterion = _('Recent usage'), lastUsageTime = None)



def run():
    """
        LD main function
    """
    global recentFilters, recentLocations, favouriteLocations, sortingCriterionForLocation
    global objectUsageCount, timeTheObjectWasLastUsed
    
    print version.announce

    # this must be before the creation of MyMainWindow
    global applic
    applic = QApplication(sys.argv)
    
    createDirIfNotExisting(config.dataDir)

    makeprograms()
    makedevices()
    makeverbs()

    # this is read by the callbacks, must be global. This is an inherent
    # problem of the callback idiom :-(
    global mainWindow
    mainWindow = MyMainWindow()
    mainWindow.setCaption("Logical Desktop")
    mainWindow.setIcon(QPixmap(config.imgDir + "italian-flag.png"))
    applic.setMainWidget(mainWindow)
    
    global locationToolbar
    locationToolbar = QToolBar(mainWindow)
    
    global bookmarksButton
    bookmarksButton = FlashingButton(QIconSet(QPixmap(config.imgDir +"gnome-favorites.png")),
                     _("Important Locations"),
                     locationToolbar)
    bookmarksButton.setMinimumHeight(50)
    
    global favouriteLocPopupMenu
    favouriteLocPopupMenu = MyPopupMenu(bookmarksButton)
    QObject.connect(favouriteLocPopupMenu, PYSIGNAL("itemActivated"), onLocationSelectedFromPopupMenu)
    bookmarksButton.setPopup(favouriteLocPopupMenu)

    global recentLocationsButton
    recentLocationsButton = FlashingButton(QIconSet(QPixmap(config.imgDir +"clock.png")),
                           _("Recent Locations"), locationToolbar)
    recentLocationsButton.setMinimumHeight(50)
    
    global recentLocPopupMenu
    recentLocPopupMenu = MyPopupMenu(recentLocationsButton)
    recentLocationsButton.setPopup(recentLocPopupMenu)
    QObject.connect(recentLocPopupMenu, PYSIGNAL("itemActivated"), onLocationSelectedFromPopupMenu)

    global executeToolbar
    executeToolbar = QToolBar(mainWindow)
    
    global executeButton
    executeButton = FlashingButton(QIconSet(), "no text yet" , executeToolbar)
    executeToolbar.hide()
    executeButton.setMinimumHeight(50)
    QObject.connect(executeButton, SIGNAL("clicked()"), onExecuteButtonPressed)
    spacing = 4

    mainVBox = QVBox(mainWindow)
    mainVBox.show()
    mainVBox.setMargin(spacing)
    mainWindow.setCentralWidget(mainVBox)

    selectedHBox = QHBox(mainVBox)
    selectedHBox.setFrameStyle(QFrame.StyledPanel | QFrame.Sunken)
    selectedHBox.setPaletteBackgroundColor(selectedHBox.palette().active().mid())
    selectedHBox.setMargin(3)
    selectedHBox.setSpacing(8)
    selectedHBox.show()

    global selectedLabel
    selectedLabel = FlashingLabel('This text will change', selectedHBox)
    selectedLabel.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
    selectedLabel.show()

    global deselectOneItemButton
    deselectOneItemButton = FlashingButton(QIconSet(QPixmap(config.imgDir + "glasses.png")), _("See selected"), selectedHBox)
    deselectOneItemButton.setMinimumHeight(32)
    deselectOneItemButton.show()
    deselectOneItemButton.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

    global selectedItemsPopupMenu
    selectedItemsPopupMenu = MyPopupMenu(deselectOneItemButton)
    deselectOneItemButton.setPopup(selectedItemsPopupMenu)

    global deselectAllButton
    deselectAllButton = QPushButton(QIconSet(QPixmap(config.imgDir + "cross.png")), _("Deselect all"), selectedHBox)
    deselectAllButton.setMinimumHeight(32)
    deselectAllButton.show()
    deselectAllButton.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
    QObject.connect(deselectAllButton, SIGNAL("clicked()"), onDeselectAllPressed)

    mainGrid = QGrid(4, mainVBox)
    mainGrid.setSpacing(spacing)
    mainGrid.setMargin(spacing)
    mainGrid.show()

    hBoxSpacing = 3

    verbHBox = QHBox(mainGrid)
    verbHBox.setSpacing(spacing)
    verbHBox.show()

    verbIcon = QLabel(verbHBox)
    verbIcon.setPixmap(QPixmap(config.imgDir + "person.png"))
    verbIcon.show()
    verbIcon.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

    verbLabel = QLabel('Verbs', verbHBox)
    verbLabel.show()

    # create three hBox, indexed by panel
    for k in ['file', 'device', 'program']:
        hBox[k] = QHBox(mainGrid)
        hBox[k].hide()
        hBox[k].setSpacing(hBoxSpacing)

        icon[k] = QLabel(hBox[k])
        icon[k].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        icon[k].show()

        locationButton[k] = WrappingButton('text will change', hBox[k], data = None)
        QObject.connect(locationButton[k], PYSIGNAL("pressed"), onLocationButtonPressed)
        QObject.connect(locationButton[k], PYSIGNAL("released"), onLocationButtonReleased)
        locationButton[k].show()

        closeButton[k] = QPushButton(QIconSet(QPixmap(config.imgDir + "delete.png")), '', hBox[k])
        closeButton[k].setFixedWidth(32)
        closeButton[k].setMinimumHeight(32)
        closeButton[k].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)
        closeButton[k].show()
    QObject.connect(closeButton['file'], SIGNAL("clicked()"), onFileCloseButtonPressed)
    QObject.connect(closeButton['program'], SIGNAL("clicked()"), onProgramCloseButtonPressed)
    QObject.connect(closeButton['device'], SIGNAL("clicked()"), onDeviceCloseButtonPressed)
        
    # SECOND ROW
    listBox['verb'] = MyListBox(frameVisible = False, parent = mainGrid)
    listBox['verb'].show()
    QObject.connect(listBox['verb'], PYSIGNAL("itemClicked"), onVerbClicked)

    for k in ["file", "device", "program"]:
        listBox[k] = MyListBox(frameVisible = False, parent = mainGrid)
        listBox[k].hide()
        QObject.connect(listBox[k], PYSIGNAL("itemDoorClicked"), onEnterLocationButtonClicked)
        QObject.connect(listBox[k], PYSIGNAL("itemClicked"), onObjectItemClicked)

    # THIRD ROW
    for k in ["verb", "file", "device", "program"]:
        filterBox[k] = QHBox(mainGrid)
        filterBox[k].hide()
        filterBox[k].setSpacing(3)

        filterClearButton[k] = QPushButton(filterBox[k])
        filterClearButton[k].setIconSet(QIconSet(QPixmap(config.imgDir + 'cross.png')))
        filterClearButton[k].setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Minimum)

        filterQuickText[k] = MyLineEdit(filterBox[k])
        QObject.connect(filterQuickText[k], PYSIGNAL("timer"), onFilterKeyTimer)
        QObject.connect(filterQuickText[k], SIGNAL("returnPressed()"), onEnterPressedOnFilter)

        filterButton[k] = QPushButton(QIconSet(QPixmap(config.imgDir + "mbuto.png")), _("Recent filters"), filterBox[k])
        filterButton[k].show()

        filterMenu[k] = MyPopupMenu(filterButton[k])
        QObject.connect(filterMenu[k], PYSIGNAL("itemActivated"), onRecentFilterClicked)
        filterButton[k].setPopup(filterMenu[k])

    QObject.connect(filterClearButton['file'], SIGNAL("clicked()"), clearFileFiltersAndUpdate)
    QObject.connect(filterClearButton['device'], SIGNAL("clicked()"), clearDeviceFiltersAndUpdate)
    QObject.connect(filterClearButton['program'], SIGNAL("clicked()"), clearProgramFiltersAndUpdate)
    QObject.connect(filterClearButton['verb'], SIGNAL("clicked()"), clearVerbFiltersAndUpdate)
        
    filterBox['verb'].show()

    # FOURTH ROW
    class SortingCriterion:
        def __init__(self, name):
            self.name = name
        def toString(self):
            return self.name

    for k in ["verb", "file", "device", "program"]:
        sortBox[k] = MyComboBox(mainGrid)
        sortBox[k].setMaximumHeight(25)
        sortBox[k].hide()
        sortBox[k].insertItem(config.imgDir + "name.png", text = _("Name"), data = "Name")
        sortBox[k].insertItem(config.imgDir + "stock_timer.png", _("Recent usage"), "Recent usage")
        sortBox[k].insertItem(config.imgDir + "favourites-small.png", _("Frequent usage"), "Frequent usage")
        sortBox[k].setCurrentText(_("Recent usage"))
        QObject.connect(sortBox[k], PYSIGNAL("itemActivated"), onSortBoxItemActivated)
    sortBox['verb'].show()
    sortBox['file'].insertItem(config.imgDir + "size.png", _("Size"), "Size")
    sortBox['file'].insertItem(config.imgDir + "clock-small.png", _("Date"), "Date")
    
    updateVisibilityOfSelectedItemsBar()
    mainWindow.showMaximized()

    redrawVerbs(verbs, selectedVerbs, selectedObjects,  centeringCommand = None)
    rebuildRecentLocationsMenu(recentLocations)
    rebuildImportantLocationsMenu(favouriteLocations, allLocation)

    verifyThatNoVerbContainsInvalidCharacters()

    dataDir = config.dataDir
    # BEGIN read database
    recentLocations2 = persistence.loadVariableFromFile(dataDir + 'recentLocations', default = [])
    recentLocations = [l for l in map(fromFileRepresentationToLocation, recentLocations2) if l != None]
    rebuildRecentLocationsMenu(recentLocations)

    favouriteLocations2 = persistence.loadVariableFromFile(dataDir + 'favouriteLocations', default = [])
    favouriteLocations = [l for l in map(fromFileRepresentationToLocation, favouriteLocations2) if l != None]
    rebuildImportantLocationsMenu(favouriteLocations, allLocation)

    timeTheObjectWasLastUsed = persistence.loadVariableFromFile(dataDir + 'timeTheObjectWasLastUsed', default = {})

    recentFilters = persistence.loadVariableFromFile(dataDir + 'recentFilters', default = {'file': [], 'device':[], 'program':[], 'verb':[]})
    rebuildFilterMenus()

    sortingCriterionForLocation = persistence.loadVariableFromFile(dataDir + 'sortingCriterionForLocation', default = {})

    objectUsageCount = persistence.loadVariableFromFile(dataDir + 'objectUsageCount', default = {})
    # END

    selectedLabel.startFlashing(config.gFlashingTimeForText, resetOriginalColorsWhenFinished = True)
    applic.exec_loop()

    # BEGIN write database
    recentLocations2 = [l.persistentID() for l in recentLocations] 
    persistence.dumpToFile(targetFile = dataDir + 'recentLocations', var = recentLocations2)
    favouriteLocations2 = [l.persistentID() for l in favouriteLocations] 
    persistence.dumpToFile(targetFile = dataDir + 'favouriteLocations', var = favouriteLocations2)
    persistence.dumpToFile(targetFile = dataDir + 'objectUsageCount', var = objectUsageCount)
    persistence.dumpToFile(targetFile = dataDir + 'recentFilters', var = recentFilters)
    persistence.dumpToFile(targetFile = dataDir + 'timeTheObjectWasLastUsed', var = timeTheObjectWasLastUsed)
    persistence.dumpToFile(targetFile = dataDir + 'sortingCriterionForLocation', var = sortingCriterionForLocation)

    for v in verbs:
        v.saveState()
    for v in programs:
        v.saveState()
    # END

if __name__ == "__main__":
    run()

